/* GWS.C: The Ghetto Webserver. 
 *
 * An extremely quick and dirty port listener/partial webserver that is
 * *not* HTTP-1.0 compliant. Also not intended to handle multiple connections
 * at once.
 *
 * Listens for an incoming TCP/IP connection on port 8080, waits for input
 * that starts with "GET" and ends with * two "\r\n" sequences, then sends
 * the file index.html in the current directory to the requester, regardless
 * of what file they actually requested. If index.html is not found, sends
 * a very simple 404 page.
 *
 *  2006/02/06, bwc - Initial
 *  2006/02/07, bwc - Header handling debugged and slightly improved.
 *  2006/02/08, bwc - Thanks to Alex Belits for linux compile and typo patch.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h> 
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/uio.h>
#include <unistd.h>



#define LISTEN_PORT 8080
#define NETBUFF_SIZE 4096
#define HTTPREQ_MAXSIZE NETBUFF_SIZE

#define HTML_FILENAME "index.html"

#define HTML_NOTFOUND "<html><head>404 Yo' Momma!</head>\
                       <body><h1>404 - Not Found</h1>\
                       <p>The Ghetto Webserver could not find the file\
                       you're asking for. Maybe it's not here?</p>\
                       </body></html>"
                         
#define HTML_BADREQUEST "<html><head>400 Whatchu talking 'bout Willis?</head>\
                         <body><h1>400 - Bad Request</h1>\
                         <p>The Ghetto Webserver could not parse your\
                         request. Sorry.</p>\
                         </body></html>"

/* Forward decls */

int getHTTPReq(int nfd);
int sendFile(int nfd);



/** MAIN **/

int main(void)
{
  struct sockaddr_in listenSock, newConnSock;
  int listenFd, errVal, newConnFd, newConnSockSize;

  /* Your typical listen(), then accept(), send data, close() loop. Since
   * design intent was to not handle multiple connections at once, I
   * will not be fork()ing off a new process for each connection. Good
   * thread management is very important for a real webserver, but is
   * also too long to include here.
   */

  /* Gotta have a socket before you can do anything else... */

  listenFd = socket(AF_INET, SOCK_STREAM, 0);
  if(listenFd < 0) {
    perror("socket() failed");
    exit(1);
  }
 
  /* bind() the socket to the webserver's port. */

  /* Wipe struct clean. */
  memset((void *)&listenSock, 0, sizeof(struct sockaddr_in));

  /* Fill in some necessary info. */
  listenSock.sin_family = AF_INET;           /* Gotta be TCP/IP */
  listenSock.sin_addr.s_addr = INADDR_ANY;   /* Accept from any IP address */
  listenSock.sin_port = htons(LISTEN_PORT);  /* Only accept on LISTEN_PORT */

  errVal = bind(listenFd, (struct sockaddr *)&listenSock, sizeof(listenSock));
  if(errVal < 0) {
    perror("Can't bind()");
    exit(2);
  }
  
  /* Tell the TCP/IP stack to listen() for incoming connections.
   * Max backlog of 5 conns. */

  errVal = listen(listenFd, 5);
  if(errVal < 0) {
    perror("Can't listen()");
    exit(3);
  }

  /** MAIN LOOP **/

  while(1) {

    newConnSockSize = sizeof(newConnSock);
    newConnFd = accept(listenFd, (struct sockaddr *)&newConnSock,
                       (socklen_t *)&newConnSockSize);

    if(newConnFd > 0) {
      if(getHTTPReq(newConnFd) != 0) {
        write(newConnFd, HTML_BADREQUEST, strlen(HTML_BADREQUEST));
      } else {
        if(sendFile(newConnFd) != 0) {
          write(newConnFd, HTML_NOTFOUND, strlen(HTML_NOTFOUND));
        }
      }
      shutdown(newConnFd, SHUT_RDWR);   /* We mean it, conn is going away! */
      close(newConnFd);

    } else {
      /* Weird non-connection. Log and discard, then wait for next. */
      perror("Non-fatal error on accept()");
    }

  } /* while(1) */

  exit(0);

} /** MAIN **/



/* sendFile(): Sends index.html to the given network file descriptor.
 *
 * params:
 *  nfd - An integer representing a file descriptor attached to a
 *    network connection, gotten from the accept() system call.
 *
 * returns: 0 if file sent, -1 if can't open file.
 */
int sendFile(int nfd)
{
  FILE *htmlFile;
  char buf[NETBUFF_SIZE];
  char *bytesRead;


  htmlFile = fopen(HTML_FILENAME, "r");
  if(NULL == htmlFile) {
    return(-1);
  } else {
    do {
      strcpy(buf, "");        /* fgets() may not modify buf if file at EOF */
      bytesRead = fgets(buf, NETBUFF_SIZE, htmlFile);
      write(nfd, buf, strlen(buf));
      /* printf("Read %i bytes from file, wrote them.\n", strlen(buf)); */
    } while(NULL != bytesRead);

    fclose(htmlFile);
  }
  return(0);
}



/* getHTTPReq(): Attempts to read an HTTP request from the given nfd.
 *
 * params:
 *  nfd - A file descriptor attached to a network socket.
 *
 * returns: 0 if good request received. -1 if the header
 * was invalid.
 */
int getHTTPReq(int nfd)
{
  char reqbuf[HTTPREQ_MAXSIZE];	/* Buffer for HTTP request. */
  int reqbufPos;		/* Where in the buffer we are. */
  int bytesRead;		/* How many bytes read off nfd each time. */
  char *beginSeq, *endSeq;	/* Location of req begin and end sequences. */

  /* Keep grabbing chars off nfd until we get what we want. Or not. */

  memset((void *)reqbuf, 0, HTTPREQ_MAXSIZE);
  reqbufPos = 0;
  beginSeq = endSeq = NULL;

  while(NULL == endSeq) {

    /* First, grab any incoming data off the connection (but not
     * so much that it overflows the request buffer.) This should
     * be some or all of the web browser's GET request.
     */
    bytesRead = read(nfd, (reqbuf + reqbufPos),
                     (HTTPREQ_MAXSIZE - reqbufPos) - 2);
    reqbufPos += bytesRead;

    /* If we read 0 bytes (on a blocking socket), and it wasn't because
     * of an interruption by signal, then something has gone wrong. */
    if(0 == bytesRead) {
      /* Connection was lost partway through the header. Oh well. */
      /* perror("Connection lost during HTTP request"); */
      return(-1);
    }

    /* Scan the request buffer for the start and end sequences. You'd
     * think we should only scan what's recently arrived, but that
     * could fail if the sequences come in one character at a time,
     * which could happen due to network or client weirdness.
     * I stress: This method of header parsing is NOT HTTP-1.0 compliant!
     */

    /* Don't look for the begin or end seq if we've already found it. */
    if(NULL == beginSeq) { beginSeq = strcasestr(reqbuf, "GET"); }
    if(NULL == endSeq) { endSeq = strstr(reqbuf, "\x0D\x0A\x0D\x0A"); }

    /* if(NULL != beginSeq) { printf("beginSeq found.\n"); }
    if(NULL != endSeq) { printf("endSeq found.\n"); } */

    /* HTTP requests are limited to HTTPREQ_MAXSIZE. Bigger than that,
     * and they are treated as malformed.
     */
    if((NULL == endSeq) && (reqbufPos >= (HTTPREQ_MAXSIZE - 1))) {
      return(-1);
    }

    /* Okay then, if we've found the ending sequence... */
    if(NULL != endSeq) {
      /* ...and the beginning sequence, AND they're in the correct order. */
      if((NULL != beginSeq) && (beginSeq < endSeq)) {
        return(0);              /* HTTP request is OK! */
      } else {
        /* We saw end without begin, or begin and end but in the wrong
         * order. Either way... */
        return(-1);
      }
    }

  /* If we got here, we've either sent the file and are done,
   * sent an error and are done, or need more input to find endSeq.
   * Go back to the top of the loop to check which it is. */
      
  } /* while(NULL == endSeq) */

  /* We should never get here, but just in case... */

  return(-1);
}

